/*
 Pipe Magic Tools
 Copyright (C) 2005 Michael Hanselmann
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 $Id: pmt.c,v 1.3 2005/06/14 19:45:30 michael Exp $
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include "fmt.h"
#include "pmt.h"

// Modules
#include "speed.h"
#include "throttle.h"
#include "rot13.h"
#include "rot47.h"
#include "case.h"

#define DEFAULT_BLOCKSIZE (1 << 16)

struct pmt_module_info pmt_modules[] = {
    PMT_ROT13,
    PMT_ROT47,
    PMT_SPEED,
    PMT_THROTTLE,
    PMT_TOLOWER,
    PMT_TOUPPER,
    { NULL }
};

unsigned long long sum = 0;
ssize_t blocksize = DEFAULT_BLOCKSIZE;

void usage(const char* progname) {
    char tmp_ulong[FMT_ULONG];

    write(STDERR_FILENO, "Usage: ", 7);
    write(STDERR_FILENO, progname, strlen(progname));
    write(STDERR_FILENO, " [--bs=<blocksize>]", 19);
    write(STDERR_FILENO, " <module> [module-parameters]\n", 30);
    write(STDERR_FILENO, "\n", 1);

    write(STDERR_FILENO, "Parameters:\n", 12);
    write(STDERR_FILENO,
          "    --bs=<blocksize>   Specifies how many bytes should be read at once\n", 71);
    write(STDERR_FILENO,
          "                       [default: ", 33);
    write(STDERR_FILENO, tmp_ulong, fmt_ulong(tmp_ulong, DEFAULT_BLOCKSIZE));
    write(STDERR_FILENO, "]\n", 2);

    write(STDERR_FILENO, "\n", 1);
    write(STDERR_FILENO, "Available modules:\n", 19);
    for(int i = 0; pmt_modules[i].name; ++i) {
        char tmp[20];
        write(STDERR_FILENO, tmp,
              fmt_pad(tmp, pmt_modules[i].name,
                      strlen(pmt_modules[i].name),
                      10, 20));
        write(STDERR_FILENO, " - ", 3);
        write(STDERR_FILENO, pmt_modules[i].desc,
              strlen(pmt_modules[i].desc));
        write(STDERR_FILENO, "\n", 1);
    }

    write(STDERR_FILENO, "\n", 1);
    write(STDERR_FILENO, "Use ", 4);
    write(STDERR_FILENO, progname, strlen(progname));
    write(STDERR_FILENO, " <module> --help to see its parameters.\n", 39);

    write(STDERR_FILENO, "\n", 1);
}

void usage_module(const char* module_name,
                  struct pmt_module_info* module_info) {
    write(STDERR_FILENO, "The ", 4);
    write(STDERR_FILENO, module_name, strlen(module_name));
    write(STDERR_FILENO, " module has ", 12);
    if(!module_info->parameters) {
        write(STDERR_FILENO, "no parameters.\n", 15);
    }else{
        write(STDERR_FILENO, "those parameters:\n", 19);
        for(int i = 0; module_info->parameters[i].parameter; ++i) {
            char tmp[30];
            write(STDERR_FILENO, tmp,
                  fmt_pad(tmp, module_info->parameters[i].parameter,
                          strlen(module_info->parameters[i].parameter),
                          20, 30));
            write(STDERR_FILENO, "   ", 3);
            write(STDERR_FILENO, module_info->parameters[i].desc,
                  strlen(module_info->parameters[i].desc));
            write(STDERR_FILENO, "\n", 1);
        }
    }
}

int main(int argc, char* argv[]) {
    struct pmt_module_info* module_info = NULL;
    int show_usage = 1;
    int module_argv_begin = 0;

    // Parse parameters
    for(int i = 1; i < argc; ++i) {
        if(!module_argv_begin) {
            if(!strncmp(argv[i], "--bs=", 5) && strlen(argv[i]) > 5) {
                blocksize = atol(argv[i] + 5);
            }else if(!strcmp(argv[i], "--help")) {
                show_usage = 1;
            }else if(strncmp(argv[i], "--", 2)) {
                module_argv_begin = i;
                show_usage = 0;
            }
        }else{
            if(!strcmp(argv[i], "--help")) {
                show_usage = 1;
            }
        }
    }

    // Module given
    if(module_argv_begin) {
        for(int i = 0; pmt_modules[i].name; ++i) {
            if(!strcmp(pmt_modules[i].name, argv[module_argv_begin])) {
                module_info = &pmt_modules[i];
                break;
            }
        }
        if(!module_info) {
            write(STDERR_FILENO, "Unknown module: ", 17);
            write(STDERR_FILENO, argv[module_argv_begin],
                  strlen(argv[module_argv_begin]));
            write(STDERR_FILENO, "\n", 1);
            return 1;
        }
    }

    if(show_usage) {
        if(!module_argv_begin) {
            usage(argv[0]);
        }else{
            usage_module(argv[module_argv_begin], module_info);
        }
        return 1;
    }

    int ret = 1;
    if(module_info->func_init) {
        ret = module_info->func_init(argc - module_argv_begin,
                                     &argv[module_argv_begin]);
    }

    if(ret) {
        if(blocksize > SSIZE_MAX) {
            blocksize = SSIZE_MAX;
        }

        char* buffer = (char*)malloc(blocksize);
        ssize_t len;

        do {
            len = read(STDIN_FILENO, buffer, blocksize);
            if(len < 0) break;

            sum += len;
            module_info->func(len, buffer);

            if(write(STDOUT_FILENO, buffer, len) < 0) {
                break;
            }
        } while(len > 0);

        free(buffer);

        if(module_info->func_end) {
            module_info->func_end();
        }
    }

    return 0;
}
